"
    Copyright (C) 2016-now Jan Vrany

    This code is not an open-source (yet). You may use this code
    for your own experiments and projects, given that:

    * all modification to the code will be sent to the
      original author for inclusion in future releases
    * this is not used in any commercial software

    This license is provisional and may (will) change in
    a future.
"
"{ Package: 'jv:dragonfly/c1' }"

"{ NameSpace: DragonFly }"

C1CompilerBackend subclass:#C1CompilerBackendX86_64
	instanceVariableNames:'method numArgs numVars module function asm dib prologue epilogue
		contextSetup literals literalsBaseAddr context
		contextFieldIndexCVars line fileDI moduleDI functionTyDI
		functionDI scopeDIStack typeDIs variableDIsStack'
	classVariableNames:''
	poolDictionaries:'DragonFly::C1LLVMTypes LLVMIntPredicate VMData VMOffsets
		VMConstants DragonFly::C1CompilerDebugFlags DWARFConstants'
	category:'DragonFly-C1'
!

!C1CompilerBackendX86_64 class methodsFor:'documentation'!

copyright
"
    Copyright (C) 2016-now Jan Vrany

    This code is not an open-source (yet). You may use this code
    for your own experiments and projects, given that:

    * all modification to the code will be sent to the
      original author for inclusion in future releases
    * this is not used in any commercial software

    This license is provisional and may (will) change in
    a future.

"
! !

!C1CompilerBackendX86_64 methodsFor:'accessing'!

functionNamed: name type: type
    | f |

    f := module getFunctionNamed: name.
    f isNil ifTrue:[ 
        f := module addFunctionNamed: name type: type
    ].
    ^ f

    "Created: / 17-06-2016 / 23:23:41 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

functionSEND: nArgs
    | name |

    self assert: (nArgs between: 0 and:15).
    name := '_SEND', nArgs printString.
    ^ self functionNamed: name type: (TyOBJFUNCs at: nArgs + 1).

    "Created: / 20-04-2016 / 21:39:08 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 17-06-2016 / 23:25:05 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

function__MKREALCONTEXT5
    ^ self functionNamed: '__MKREALCONTEXT5' type: TyMKREALCONTEXT5

    "Created: / 17-06-2016 / 23:25:40 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

function__SSEND: nArgs
    | name |

    self assert: (nArgs between: 0 and:15).
    name := '__SSEND', nArgs printString.
    ^ self functionNamed: name type: (TySSENDSs at: nArgs + 1).

    "Created: / 23-06-2016 / 22:04:36 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

function__debugBreakPoint3__
    ^ self functionNamed: '__debugBreakPoint3__' type: (LLVMType function: #() returning: LLVMType void)

    "Created: / 24-06-2016 / 01:04:57 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

function__interruptL
    ^ self functionNamed: '__interruptL' type: (LLVMType function: { LLVMType int32 } returning: LLVMType void)

    "Created: / 22-08-2016 / 23:22:23 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

method
    ^ method
!

method:aMethod
    method := aMethod.
    numArgs := aMethod numArgs.
    method isJavaMethod ifTrue:[ 
        numArgs := (method isStatic ifTrue:[0] ifFalse:[1])
    ].
    numVars := aMethod numVars.

    "Modified: / 07-08-2016 / 22:54:17 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!C1CompilerBackendX86_64 methodsFor:'codegen-boxing / unboxing'!

makeInteger32:anLLVMValueOrInteger
    ^ self makeSmallInteger: anLLVMValueOrInteger.

    "Created: / 06-08-2016 / 18:34:09 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

makeSmallInteger:anLLVMValueOrInteger 
    | value |

    anLLVMValueOrInteger isInteger ifTrue:[ 
        self assert:(anLLVMValueOrInteger between:SmallInteger minVal and:SmallInteger maxVal).
        value := LLVMConstant sintptr: anLLVMValueOrInteger.
    ] ifFalse:[
        | type intptrType |

        type := anLLVMValueOrInteger type.
        self assert: type isIntegerType.
        intptrType := LLVMType intptr.
        self assert: type sizeInBits <= intptrType sizeInBits.
        (type sizeInBits < intptrType sizeInBits) ifTrue:[ 
            value := asm sext: anLLVMValueOrInteger toInt: intptrType.
        ] ifFalse:[
            value := anLLVMValueOrInteger.
        ].
    ].
    ^ asm 
        int:(asm or:(asm shl: value _: (LLVMConstant uintptr: 1)) _: (LLVMConstant uintptr: 1))
        toPtr:TyOBJ

    "Created: / 18-06-2016 / 22:15:48 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 05-08-2016 / 16:59:38 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

valBoolean:anLLVMValue
    | type |

    type := anLLVMValue type.
    type isIntegerType ifTrue:[ 
        ^ type sizeInBits > 1 
            ifTrue:[asm trunc: anLLVMValue toInt: LLVMType int1]
            ifFalse:[ anLLVMValue ]
    ].
    self notYetImplemented.

    "Created: / 15-08-2016 / 23:41:34 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

valInteger32:anLLVMValue
    | v |

    v := self valSmallInteger:anLLVMValue.
    v type sizeInBits > 32 ifTrue:[ 
        v := asm trunc: v toInt: LLVMType int32.
    ].
    ^ v

    "Created: / 10-08-2016 / 23:30:57 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 11-08-2016 / 09:03:58 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

valSmallInteger:anLLVMValue
    self assert: anLLVMValue type = TyOBJ.

    ^ asm ashr: (asm ptr: anLLVMValue toInt: LLVMType intptr) _: (LLVMConstant sintptr: 1)

    "Created: / 10-08-2016 / 23:31:32 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 14-08-2016 / 09:01:30 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!C1CompilerBackendX86_64 methodsFor:'codegen-conditionals'!

emitIf: cond then: then
    ^ asm if: cond then: then

    "Created: / 15-08-2016 / 23:44:56 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

emitJumpTo: label
    asm br: label

    "Created: / 15-08-2016 / 23:49:04 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!C1CompilerBackendX86_64 methodsFor:'codegen-load / store'!

loadClassFromILC:pIlc 
    self assert:pIlc type = TyInlineCachePtr.
    ^ asm load:(asm gep:pIlc at:{ 0 . TyInlineCacheIndexClass })
             "Created: / 12-02-2016 / 13:19:19 / Jan Vrany <jan.vrany@fit.cvut.cz>" "Modified: / 15-04-2016 / 23:18:52 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

loadClassOf:obj
    | class1 class2 class3 class23|


    asm if: (self emitIsNilObject:obj) then:[ 
        class1 := self loadClassUndefinedObject.
    ] else: [ 
        asm if: (self emitIsSmallIntegerObject:obj) then: [
            class2 := self loadClassSmallInteger.
        ] else:[  
            class3 := (asm load:(asm gep:obj at:{ 0 . TyInstanceFieldIndexClass })).
        ].
        class23 := asm phi: { class2 . class3 }.
    ].
    ^ asm phi: { class1 . class23 }

    "Created: / 09-02-2016 / 17:25:48 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 18-06-2016 / 20:50:51 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

loadClassSmallInteger
    | addr |

    addr := asm 
                int:(LLVMConstant uintptr:ADDR_SmallInteger)
                toPtr:TyOBJ pointer
                as:'ADDR_SmallInteger'.
    ^ asm load:addr as: 'SmallInteger'.

    "Created: / 12-02-2016 / 11:54:10 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 21-04-2016 / 17:03:02 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

loadClassUndefinedObject
    | addr |

    addr := asm 
                int:(LLVMConstant uintptr:ADDR_UndefinedObject )
                toPtr:TyOBJ pointer
                as:'ADDR_UndefinedObject'.
    ^ asm load:addr as: 'UndefinedObject'.

    "Created: / 18-06-2016 / 00:47:48 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

loadContextVarAt: index
    index <= numArgs ifTrue:[ 
        context isNil ifTrue:[ 
            ^ function parameterAt: OBJFUNCArgIndexArgBase + ((method isJavaMethod and: [method isStatic not]) ifTrue:[-1] ifFalse:[0]) + index.
        ].
    ].
    self assert: context notNil description: 'No context!!'.
    ^ asm load: (asm gep: context at:{ 0 . contextFieldIndexCVars . index - 1 }) as: 'context@', index printString.

    "Created: / 31-07-2016 / 23:10:47 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 09-08-2016 / 11:04:33 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

loadInterruptPending
    | addr |

    addr := asm 
                int:(LLVMConstant uintptr:ADDR_InterruptPending)
                toPtr:TyOBJ pointer
                as:'ADDRInterruptPending'.
    ^ asm load:addr as: 'InterruptPending'.

    "Created: / 22-08-2016 / 23:25:08 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

loadLiteral: anObject
    | index |
    anObject class == SmallInteger ifTrue:[ 
        ^ self makeSmallInteger:anObject.  
    ].
    anObject isNil ifTrue:[ 
        ^ self loadNil.
    ].
    index := literals identityIndexOf: anObject.
    index == 0 ifTrue:[
        literals add: anObject.
        index := literals size.
    ].
    ^ asm load: (asm gep: (asm load: literalsBaseAddr) at: index - 1)
            as: ('literal_', index printString)

    "Created: / 20-06-2016 / 20:41:30 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 20-06-2016 / 23:35:48 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

loadNil
    ^ asm int:(LLVMConstant uintptr:0) toPtr:TyOBJ as: 'nil'

    "Created: / 21-04-2016 / 13:59:08 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 20-06-2016 / 23:36:50 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

loadReceiver
    ^ context isNil 
        ifTrue:[ function parameterAt: OBJFUNCArgIndexReceiver ]
        ifFalse:[ asm load: (asm gep: context at:{ 0 . TyContextFieldIndexReceiver }) as: 'zelf' ]

    "Created: / 20-04-2016 / 22:09:34 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 20-06-2016 / 23:37:09 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

loadRetvalTemp
    self assert: context notNil description: 'No context!!'.
    ^ asm load: (asm gep: context at:{ 0 . TyContextFieldIndexRetvalTemp }) as: 'retvalTemp'

    "Created: / 17-06-2016 / 23:13:50 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

loadSearchClass
    ^ context isNil 
        ifTrue:[ function parameterAt: OBJFUNCArgIndexSearchClass ]
        ifFalse:[ asm load: (asm gep: context at:{ 0 . TyContextFieldIndexSearchClass }) as: 'searchClass' ]

    "Created: / 20-04-2016 / 23:27:03 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 20-06-2016 / 23:37:24 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

loadSelector
    ^ context isNil 
        ifTrue:[ function parameterAt: OBJFUNCArgIndexSelector ]
        ifFalse:[ asm load: (asm gep: context at:{ 0 . TyContextFieldIndexSelector }) as: 'selector']

    "Created: / 20-04-2016 / 23:24:12 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 20-06-2016 / 23:37:33 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

loadThisContext
    | addr |

    addr := asm 
                int:(LLVMConstant uintptr:ADDR___thisContext)
                toPtr:TyOBJ pointer
                as:'ADDR___thisContext'.
    ^ asm load:addr as: '__thisContext'.

    "Created: / 17-06-2016 / 23:38:20 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

storeContextVar: value at: index
    self assert: context notNil description: 'No context!!'.
    ^ asm store: value at: (asm gep: context at:{ 0 . contextFieldIndexCVars . index - 1 })

    "Created: / 31-07-2016 / 23:07:38 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 07-08-2016 / 22:30:08 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

storeRetvalTemp: value
    self assert: context notNil description: 'No context!!'.
    ^ asm store: value at: (asm gep: context at:{ 0 . TyContextFieldIndexRetvalTemp })

    "Created: / 17-06-2016 / 23:14:17 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

storeThisContext: value
    | addr valueAsOBJ |

    valueAsOBJ := value type = TyOBJ 
        ifTrue:[ value ]
        ifFalse:[ asm bitcast: value to: TyOBJ ].

    addr := asm 
                int:(LLVMConstant uintptr:ADDR___thisContext)
                toPtr:TyOBJ pointer
                as:'ADDR___thisContext'.
    asm store: valueAsOBJ at: addr.

    "Created: / 17-06-2016 / 23:38:50 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!C1CompilerBackendX86_64 methodsFor:'codegen-misc'!

emitAddW: v1 with: v2
    | ty |

    ty := v1 type.
    self assert: ty isIntegerType.
    self assert: ty = v2 type.

    ^ asm extractvalue: (asm call: (module getIntrinsicNamed:'llvm.sadd.with.overflow' types: { ty }) _: { v1 . v2 })at: 0

    "Created: / 10-08-2016 / 23:26:26 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

emitDebugBreakPoint
    asm call: (self function__debugBreakPoint3__) _: #()

    "Created: / 24-06-2016 / 01:05:50 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

emitICmp: v1 with: v2 predicate: predicate
    | ty |

    ty := v1 type.
    self assert: ty isIntegerType.
    self assert: ty = v2 type.

    ^ asm icmp: v1 _: v2 cond: predicate

    "Created: / 15-08-2016 / 23:28:44 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

emitInterruptCheck
    asm if: (self emitIsNotNilObject: self loadInterruptPending) then:[ 
        self emitInterruptL: 1.
    ].

    "Created: / 22-08-2016 / 23:19:58 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

emitInterruptL: lineNr
    asm call: (self function__interruptL) _: { LLVMConstant sint32: lineNr }

    "Created: / 22-08-2016 / 23:30:37 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

emitMulW: v1 with: v2
    | ty |

    ty := v1 type.
    self assert: ty isIntegerType.
    self assert: ty = v2 type.

    ^ asm extractvalue: (asm call: (module getIntrinsicNamed:'llvm.smul.with.overflow' types: { ty }) _: { v1 . v2 })at: 0

    "Created: / 10-08-2016 / 23:27:18 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

emitReturn: retVal
    self storeRetvalTemp:  retVal.
    asm br: epilogue.

    "Created: / 23-06-2016 / 22:33:37 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

emitSubW: v1 with: v2
    | ty |

    ty := v1 type.
    self assert: ty isIntegerType.
    self assert: ty = v2 type.

    ^ asm extractvalue: (asm call: (module getIntrinsicNamed:'llvm.ssub.with.overflow' types: { ty }) _: { v1 . v2 })at: 0

    "Created: / 10-08-2016 / 23:27:06 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!C1CompilerBackendX86_64 methodsFor:'codegen-prologue / epilogue'!

emitContextSetup
    | flags arg1Index ctxSize ctxTy ctx |

    contextSetup := function addBasicBlockNamed:'context-setup'.
    asm continue:contextSetup.

    arg1Index := 1.
    ctxSize := numArgs + method numVars + method stackSize.

    method isJavaMethod ifTrue:[ 
        arg1Index := method isStatic ifTrue:[ 1 ] ifFalse:[ 2 ].
        ctxTy := C1LLVMTypes tyJContext: ctxSize.
        contextFieldIndexCVars := TyJContextFieldIndexCVars
    ] ifFalse:[method isBlock ifTrue:[
        ctxTy := C1LLVMTypes tyBContext: ctxSize.
        contextFieldIndexCVars := TyBContextFieldIndexCVars
    ] ifFalse:[
        ctxTy := C1LLVMTypes tyMContext: ctxSize.
        contextFieldIndexCVars := TyMContextFieldIndexCVars
    ]].


    (method isJavaMethod and:[ method isStatic not]) ifTrue:[ 
        ctxSize := ctxSize + 1.
    ].
    ctxTy := (method isJavaMethod ifTrue:[ C1LLVMTypes tyJContext: ctxSize ] ifFalse:[ method isBlock ifTrue:[ C1LLVMTypes tyBContext: ctxSize ] ifFalse:[ C1LLVMTypes tyMContext: ctxSize ]]).
    ctx := asm alloca:ctxTy as:'__context'.

    "/ Nil out contents...
    asm memset: ctx _: (LLVMConstant uint8: 0) _: (LLVMConstant uint64: ctx type pointee sizeInBytes) _: ctx type alignmentInBytes _: false.
    flags := (method isJavaMethod ifTrue:[ __LAZYJCON ] ifFalse:[ method isBlock ifTrue:[ __LAZYBCON ] ifFalse:[ __LAZYMCON ]]) 
            | __CANNOT_RETURN "/ For now, we don't (yet) fill setjmp() buffer                     
            | __METHOD_VALID
            | (       numArgs bitShift:__NARG_SHIFT) 
            | (method numVars bitShift:__NVAR_SHIFT) 
            | (method stackSize bitShift:__NTMP_SHIFT).
    asm store:(self makeSmallInteger:flags)
        at:(asm gep:ctx at:{ 0 . TyContextFieldIndexFlags }).    
    asm store:(self loadLiteral:method)
        at:(asm gep:ctx at:{ 0 . TyContextFieldIndexMethod }).
    asm store:self loadReceiver
        at:(asm gep:ctx at:{ 0 . TyContextFieldIndexReceiver }).
    asm store:self loadSelector
        at:(asm gep:ctx at:{ 0 . TyContextFieldIndexSelector }).
    asm store:self loadSearchClass
        at:(asm gep:ctx at:{ 0 . TyContextFieldIndexSearchClass }).
    asm store:(asm bitcast:self loadThisContext to:LLVMType intptr pointer)
        at:(asm gep:ctx at:{ 0 . TyContextFieldIndexSenderS }).
    "/ Store arguments
    1 to: method numArgs do:[:i | 
        | arg |

        arg := function parameterAt: OBJFUNCArgIndexArgBase + i.
        asm store: arg at: (asm gep: ctx at:{ 0 . contextFieldIndexCVars . i + arg1Index - 2 })      
    ].
    self storeThisContext:ctx.
    context := ctx.
    DebugContext ifTrue:[
        self emitDebugBreakPoint
    ].

    "Created: / 20-04-2016 / 23:12:22 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 07-08-2016 / 23:17:58 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

emitEpilogue
    asm continue:epilogue.
    asm ret:(asm call:self function__MKREALCONTEXT5
                _:{
                        self loadRetvalTemp
                    })
!

emitPrologue
    | classOfReceiver  classInILC  classCheck |

    prologue := function addBasicBlockNamed:'prologue'.
    epilogue := function addBasicBlockNamed:'epilogue'.
    asm continue:prologue.
    classOfReceiver := self 
            loadClassOf:(function parameterAt:OBJFUNCArgIndexReceiver).
    classInILC := self 
            loadClassFromILC:(function parameterAt:OBJFUNCArgIndexILC).
    classCheck := asm 
            icmp:(asm ptr:classOfReceiver toInt:LLVMType intptr)
            _:(asm ptr:classInILC toInt:LLVMType intptr)
            cond:LLVMIntNE.
    asm if:classCheck
        then:[
            "/ Class check failed, call _SENDX()"
            asm ret:(asm call:(self functionSEND:method numArgs)
                        _:{
                                function parameterAt:OBJFUNCArgIndexReceiver.
                                function parameterAt:OBJFUNCArgIndexSelector.
                                function parameterAt:OBJFUNCArgIndexSearchClass.
                                function parameterAt:OBJFUNCArgIndexILC
                            } , (1 to:method numArgs
                                            collect:[:i | function parameterAt:OBJFUNCArgIndexArgBase + i ])).
        ].
    self emitContextSetup.
    asm continue:(function addBasicBlockNamed:'body').

    "Created: / 09-02-2016 / 17:07:28 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 24-06-2016 / 00:05:23 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!C1CompilerBackendX86_64 methodsFor:'codegen-sends'!

emitSend: selector to: receiver with: arguments
    self assert: (arguments size between: 0 and: 15) description: 'Invalid number of arguments. VM supports max 15 args'.

    ^ asm call: (self function__SSEND: arguments size) _: {
        receiver .                  "/ receiver
        self loadLiteral: selector ."/ selector
        LLVMConstant sint32: 0.     "/ lineNr (no line info for now)
        } , arguments

    "Created: / 23-06-2016 / 22:17:35 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!C1CompilerBackendX86_64 methodsFor:'codegen-tests'!

emitIsNilObject:value 
    self assert:value type = TyOBJ.
    ^ asm 
        icmp:(asm ptr:value toInt:LLVMType intptr)
        _:(LLVMConstant uintptr:0)
        cond:LLVMIntEQ

    "Created: / 18-06-2016 / 00:48:37 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

emitIsNotNilObject:value 
    self assert:value type = TyOBJ.
    ^ asm 
        icmp:(asm ptr:value toInt:LLVMType intptr)
        _:(LLVMConstant uintptr:0)
        cond:LLVMIntNE

    "Created: / 22-08-2016 / 23:28:05 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

emitIsSmallIntegerObject:value 
    self assert:value type = TyOBJ.
    ^ asm 
        icmp:(asm and:(asm ptr:value toInt:LLVMType intptr)
                _:(LLVMConstant uintptr:1))
        _:(LLVMConstant uintptr:1)
        cond:LLVMIntEQ

    "Created: / 09-02-2016 / 17:30:55 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 12-02-2016 / 11:55:11 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!C1CompilerBackendX86_64 methodsFor:'compilation'!

finish
    "Finish the compilation, i.e., generate machine code
     and update method's code pointer"
    | opts mm object jit literalsSection |

    object := CompiledCodeObject forCompiledCode: method text: 0 literals: literals size ilcs: 0.
    literals notEmptyOrNil ifTrue:[ 
        literalsSection := object literals.
        1 to: literals size do:[:i | 
            literalsSection at: i put: (literals at: i).
        ].
        literalsBaseAddr initializer: (LLVMConstant pointer: literalsSection address type: TyOBJVec)
    ].
    dib finish.
    module verify.
    mm := C1LLVMMCJITMemoryManager for: object.
    opts := LLVMMCJITCompilerOptions new.
    opts MCJMM: mm.
    jit := LLVMExecutionEngine newForModule: module options: opts.
    "/ Sigh, for now, we have to prevent LLVMExecutionEngine instance from
    "/ being GC'd because when it is, debug symbols are gone.
    "/ 
    "/ In a long term, we should extract symfile from LLVM and register/deregister it
    "/ outselfs..
    method isJavaMethod ifTrue:[ 
        method at: 1 put: jit.
    ] ifFalse:[ 
        "/ Not yet supported - must find a way to keep the reference
        "/ in other methods...
        self notYetImplemented
    ].
    method code: (jit addressOfFunction: function).

    "Created: / 21-04-2016 / 09:19:12 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 26-08-2016 / 14:23:14 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

prepare
    | name source |
    name := self class functionNameForClass: method mclass ? UndefinedObject selector: method selector ? '<unbound>'.
    module := LLVMModule newWithName: name.
    function := module addFunctionNamed: name type: (TyOBJFUNCs at: method numArgs + 1).
    (function parameterAt: OBJFUNCArgIndexReceiver) name: 'zelf'.
    (function parameterAt: OBJFUNCArgIndexSelector) name: 'selector'.
    (function parameterAt: OBJFUNCArgIndexSearchClass) name: 'searchClass'.
    (function parameterAt: OBJFUNCArgIndexILC) name: 'pilc'.
    1 to: method numArgs do:[:i | (function parameterAt: OBJFUNCArgIndexArgBase + i) name: 'arg' , i printString ].
    asm := function builder.
    literals := OrderedCollection with: method.
    literalsBaseAddr := module addGlobalNamed: '__literals' value: (LLVMConstant null: TyOBJVec).

    dib := module debugInfoBuilder.
    source := method mclass sourceStream.
    [  
        self assert: source notNil.
        self assert: source isFileStream.
        source isFileStream ifTrue:[
            | sourceFile arg1Index |

            sourceFile := source asFilename asAbsoluteFilename.
            fileDI := dib createFile: sourceFile.
            moduleDI := dib createCompilationUnit: sourceFile language: (method isJavaMethod ifTrue:[DW_LANG_Java] ifFalse:[DW_LANG_lo_user + 10]) producer: 'DragonFly/C1'.
            functionTyDI := dib createTypeFunction: #().
            functionDI := dib createFunction: name in: fileDI file: fileDI  line: method sourceLineNumber type: functionTyDI local: true definition: true optimized: false.
            scopeDIStack := Stack with: functionDI.
            arg1Index := (method isJavaMethod and:[ method isStatic not ]) ifTrue:[ 2 ] ifFalse:[ 1 ].
            variableDIsStack := Stack with: (Array new: arg1Index - 1 + method numArgs + method numVars).
            function metadataAt: 0 put: functionDI.
        ]
    ] ensure: [ source close ].
    typeDIs := Dictionary new.

    "Created: / 21-04-2016 / 09:17:06 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 30-08-2016 / 17:04:27 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 21-10-2016 / 11:26:54 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!C1CompilerBackendX86_64 methodsFor:'debug info'!

diAssignVariableAt: index with: value
    "Mark variable at given frame index as being assigned given value"

    | variableDI  |

    variableDI := variableDIsStack top at: index + 1.
    self assert: variableDI notNil description: 'Variable at given index was not declared. Missing call to #diDeclareVariable:at:?'.
    dib insertValue: value variable: variableDI expression: dib createExpression location: asm location atEndOf: asm block offset: 0

    "Created: / 22-09-2016 / 23:22:23 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

diDeclareVariable: name at: index
    "Declares variable name at frame index `index` in current scope."

    | variableDI variableAddr |

    variableDI := dib createAutomaticVariable: name in: scopeDIStack top file: fileDI line: line type: self diTypeOBJ.
    variableDIsStack top at: index + 1 put: variableDI.
    variableAddr := asm gep: context at:{ 0 . contextFieldIndexCVars . index - 1 }.
    dib insertDeclare: variableAddr variable: variableDI expression: dib createExpression location: asm location atEndOf: asm block

    "Created: / 22-09-2016 / 22:16:44 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

diLine: line
    ^ self diLine: line column: 1

    "Created: / 23-08-2016 / 22:41:28 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

diLine: ln column: column
    line := ln.
    asm line: ln column: column scope: scopeDIStack top

    "Created: / 23-08-2016 / 22:41:54 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 22-09-2016 / 22:19:12 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

diTypeOBJ
    "Returns a DI for TyOBJ"

    ^ typeDIs at: TyOBJ ifAbsentPut:[ 
        dib createTypeScalar: 'OBJ' type: LLVMType intptr encoding: DW_ATE_unsigned.
    ].

    "Created: / 22-09-2016 / 21:49:39 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 21-10-2016 / 11:27:04 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!C1CompilerBackendX86_64 methodsFor:'labels'!

createLabel: aString
    ^ function addBasicBlockNamed: aString

    "Created: / 15-08-2016 / 23:06:32 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

placeLabel: anObject
    asm continue: anObject.

    "Created: / 15-08-2016 / 23:07:02 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!C1CompilerBackendX86_64 class methodsFor:'documentation'!

version_HG

    ^ '$Changeset: <not expanded> $'
! !

